/******************************************************************************
*
* UserTextWindow   v 0 . 9 0    Beta 1
* Copyright  2004 David Randall
*
*******************************************************************************
*
* UserTextWindow creates fast scrollable text windows with word wrapping.
*
* UserTextWindow requires TIGCC 0.95 (0.94 ??) and Extgraphv v2.00 to compile.
* UserTextWindow requires user defined font files.
*
* UserTextWindow is meant to be an extension to the Extgraph API by providing
* very fast text output to the screen.
* 
* Thanks to the TIGCC and Extgraph teams. They have made programming our
* calculators fast, friendly, and fun.
*
* Send comments to drandall99@charter.net
*
******************************************************************************/

#include <tigcclib.h>
#include "extgraph.h"
#include "UserTextWindow.h"


void UTW_Create(
		unsigned char topLeftX, 
		unsigned char topLeftY, 
		unsigned char width, 
		unsigned char height, 
		unsigned char margin,
		unsigned char lineSpacing,
		unsigned char autoScroll,
		unsigned char wordWrap,
		unsigned short textColor,
		unsigned short bkgdColor,
		unsigned char* pFont, 
		unsigned char fontWidth, 
		unsigned char fontHeight,
		unsigned char bufferLines,
		unsigned long pCurrentPlane0,
		unsigned long pCurrentPlane1,
		UserTextWindowType* utw)
{
	if ((topLeftX + width  >= 240) ||
			(topLeftY + height >= 128) ||
			((margin * 2) + fontWidth  >= width)  ||
			((margin * 2) + fontHeight >= height) ||
			(fontWidth > 8) ||
			(textColor > 3) || 
			(bkgdColor > 3) )
	{
		// This UTW is not valid
		utw->valid = 0;
	}
	else
	{
		utw->valid = 1; // Mark this UTW as valid
		utw->autoScroll = autoScroll;
		utw->wordWrap = wordWrap;
		utw->leftX = topLeftX;
		utw->rightX = topLeftX + width;
		utw->topY = topLeftY;
		utw->bottomY = topLeftY + height;
		utw->printLeftX = topLeftX + margin;
		utw->printRightX = utw->rightX - margin;
		utw->printTopY = topLeftY + margin;
		utw->printBottomY = utw->bottomY - margin + 1;
		utw->printWidth = utw->printRightX - utw->printLeftX;
		utw->lineSpacing = lineSpacing;
		utw->textColor = textColor;
		utw->bkgdColor = bkgdColor;
		utw->pFont = pFont;
		utw->fontWidth = fontWidth;
		utw->fontHeight = fontHeight;
		utw->pCurrentPlane0 = pCurrentPlane0;
		utw->pCurrentPlane1 = pCurrentPlane1;
		utw->maxCharsPerLine = (utw->printWidth + 1) / utw->fontWidth;
		utw->maxVisibleLines = (utw->printBottomY - utw->printTopY + 1) / (utw->fontHeight + utw->lineSpacing);
		
		// Set the number of text lines in the buffer 
		if (bufferLines > utw->maxVisibleLines)
			utw->maxBufferLines = bufferLines;
		else
			utw->maxBufferLines  = utw->maxVisibleLines;
		
		// Allocate the memory for the text buffer
		utw->pBuffer = malloc(utw->maxBufferLines * utw->maxCharsPerLine);
		if (!utw->pBuffer)
			utw->valid = 0;	// Not enough memory, mark the UTW as invalid
		
		// Clear the window and set the cursor position to home
		UTW_ClearHome(utw);	
	}
}

void UTW_Close(UserTextWindowType* utw)
{
	// Ensure the UTW is valid
	if (utw->valid)
	{
		// De-allocate the memory that was allocated in UTW_Create()
		free (utw->pBuffer);
		// Mark as invalid;
		utw->valid = 0;
	}
}

void UTW_ClearHome(UserTextWindowType* utw)
{
	// Ensure the UTW is valid
	if (utw->valid)
	{
		// Clear the entire text buffer
		memset(utw->pBuffer, 0, utw->maxBufferLines * utw->maxCharsPerLine);
		
		// Initialize necessary variables
		utw->scrollableLineIndex  = 0;
		utw->viewableLineIndex  = 0;
		utw->writeLineIndex = 0;
		utw->lineCount      = 0;
		
		// Clear the window
		UTW_ClearScreen(utw);
	}
}

void UTW_ClearScreen(UserTextWindowType* utw)
{
	// Ensure the UTW is valid
	if (utw->valid)
	{
		// Fill the UTW with the specified color
		FastFillRect(*(long*)(utw->pCurrentPlane0),utw->leftX,utw->topY,utw->rightX,utw->bottomY,utw->bkgdColor&0x0001);
		FastFillRect(*(long*)(utw->pCurrentPlane1),utw->leftX,utw->topY,utw->rightX,utw->bottomY,(utw->bkgdColor>>1)&0x0001);
		
		// Set the cursor to the top left corner
		utw->cursorX = utw->printLeftX;
		utw->cursorY = utw->printTopY;
	}
}

void UTW_Print(UserTextWindowType* utw, char* str)
{
	// Ensure the UTW is valid
	if (utw->valid)
	{
		char* tempStr1 = str - 1;
		char* tempStr2;
		char* lineOfText;
		char nextChar;
		unsigned char i;
		unsigned char charStart;
		unsigned char spaceIndex;
		unsigned char charCount;
		unsigned char charIndex = 0;
		unsigned char done = 0;
		
		while(!done)
		{
			charStart  = charIndex;
			spaceIndex = 0;
			charCount  = 0;
			
			while (1)
			{
				tempStr1++;
				charIndex++;
				charCount++;
			
				if (!(*tempStr1))
				{
					// We have reach the end of the string
					done = 1;
					break;
				}
				else if (*tempStr1 == 10)
				{
					// This is a carriage return with a line feed
					break;
				}
				else if (*tempStr1 == 13)
				{
					// This is a carriage return with no line feed
					// Decrement the write pointer it will be incremented later
					utw->writeLineIndex = (--utw->writeLineIndex) % utw->maxBufferLines;	
					break;
				}
				else if (*tempStr1 == 32)
				{
					// This is a space. Save the location of this space for word wrapping
					spaceIndex = charIndex;
				}
				
				if (utw->wordWrap)
				{
					// Word wrapping enabled for this UTW
					if (charCount == utw->maxCharsPerLine)
					{
						// We are at the end of the visible printing area
						// Get the next character
						nextChar = *(tempStr1+1);
						
						// Is the next character a space or a carriage return?
						if (nextChar == 0x20 || nextChar == 0x0d || nextChar == 0x0a)
						{
							// skip over these characters for the start of the next line
							tempStr1++;
							charIndex++;
						}
						else if (spaceIndex)
						{
							// Go back to the last space within the current line
							tempStr1 -= (charIndex - spaceIndex);
							charCount = spaceIndex - charStart;
							charIndex = spaceIndex;
						}
						
						break;
					}
				}
			}
			
			// Get the address of the next buffer
			lineOfText = utw->pBuffer + (utw->writeLineIndex * utw->maxCharsPerLine);
			// Increment the write pointer
			utw->writeLineIndex = (++utw->writeLineIndex) % utw->maxBufferLines;
			// Get the start address of the string to put in the buffer
			tempStr2 = str + charStart;
			
			// Limit the number of characters per line
			if (charCount > utw->maxCharsPerLine)
				charCount = utw->maxCharsPerLine;
			
			// Copy the a portion of the string to the buffer	
			for (i = 0; i < charCount; i++)
				*lineOfText++ = *tempStr2++;
			
			// Null terminate the string just in case
			if (i != utw->maxCharsPerLine)
				*lineOfText = 0;
			
			if (utw->autoScroll)
			{		
				// Auto scrolling is enabled for this UTW?
				// Increment the number of text lines added to the window
				utw->lineCount++;
				
				// Should we move the index marking the beginning of the current viewable text?	
				if (utw->lineCount > utw->maxVisibleLines)
					utw->viewableLineIndex = (++utw->viewableLineIndex) % utw->maxBufferLines;
				
				// When the lines of text begin to wrap within the buffer,
				// adjust the start index of the entire scrollable window area.
				if (utw->lineCount > utw->maxBufferLines)
					utw->scrollableLineIndex = (++utw->scrollableLineIndex) % utw->maxBufferLines;
				
				// Once the line count has exceeded the maximum text lines in the buffer, 
				// we will always be at the end of the scrollable text area.	
				if (utw->lineCount >= utw->maxBufferLines)
					utw->scrollLimit = 1;
			}
		}
	} 
}

void UTW_Draw(UserTextWindowType* utw)
{
	// Ensure the UTW is valid
	if (utw->valid)
	{
		char* tempStr;
		unsigned char i;
		unsigned char currentLine = 0;
		// A clever method to determine which Sprite8 routine to use for drawing the character
		unsigned short colorCombo = utw->textColor ^ utw->bkgdColor;
		
		UTW_ClearScreen(utw);
		
		// Print all visible lines of the window
		while (currentLine < utw->maxVisibleLines)
		{
			// Get the start address of the current line of text from the buffer
			tempStr = utw->pBuffer + (((utw->viewableLineIndex + currentLine++) % utw->maxBufferLines) * utw->maxCharsPerLine);
			
			if (colorCombo == 0)
			{
				// Dk Gray on Dk Gray  2^2=0
				// White on White      0^0=0
				// Black on Black      3^3=0
				// Lt Gray on Lt Gray  1^1=0
				
				// This text is not visible. Do nothing.
			}
			else if (colorCombo == 1)
			{
				// Dk Gray on Black    2^3=1
				// White on Lt Gray    0^1=1
				// Black on Dk Gray    3^2=1 
				// Lt Gray on White    1^0=1
	
				for (i = 0; i < utw->maxCharsPerLine; i++)
				{
					if (!(*tempStr))
						break;
	
					Sprite8_XOR_R(
							utw->cursorX, 
							utw->cursorY, 
							utw->fontHeight, 
							utw->pFont + (*tempStr++ * utw->fontHeight), 
							*(long*)(utw->pCurrentPlane0));
					utw->cursorX += utw->fontWidth;
				}
			}
			else if (colorCombo == 2)
			{
				// White on Dk Gray    0^2=2
				// Lt Gray on Black    1^3=2
				// Black on Lt Gray    3^1=2
				// Dk Gray on White    2^0=2
	
				for (i = 0; i < utw->maxCharsPerLine; i++)
				{
					if (!(*tempStr))
						break;
		
					Sprite8_XOR_R(
							utw->cursorX, 
							utw->cursorY, 
							utw->fontHeight, 
							utw->pFont + (*tempStr++ * utw->fontHeight), 
							*(long*)(utw->pCurrentPlane1));
					utw->cursorX += utw->fontWidth;
				}
			}
			else if (colorCombo == 3)
			{
				// White on Black      0^3=3
				// Black on White      3^0=3
				// Dk Gray on Lt Gray  2^1=3
				// Lt Gray on Dk Gray  1^2=3
	
				for (i = 0; i < utw->maxCharsPerLine; i++)
				{
					if (!(*tempStr))
						break;
	
					GraySingleSprite8_XOR_R(
							utw->cursorX, 
							utw->cursorY, 
							utw->fontHeight, 
							utw->pFont + (*tempStr++ * utw->fontHeight), 
							*(long*)(utw->pCurrentPlane0), 
							*(long*)(utw->pCurrentPlane1));
					utw->cursorX += utw->fontWidth;
				}
			}
			
			// This is a carriage return and a line feed
			utw->cursorX  = utw->printLeftX;
			utw->cursorY += utw->fontHeight + utw->lineSpacing;
		}
	} 
}

short UTW_Scroll(UserTextWindowType* utw, signed char lines)
{
	// Ensure the UTW is valid
	if (utw->valid)
	{
		if (!utw->scrollLimit)
		{
			signed char i = 0;
			
			if (lines > 0)
			{
				// user wants to scroll the window up
				while (i < lines)
				{
					i++;
					
					if ((utw->viewableLineIndex + utw->maxVisibleLines + i) % utw->maxBufferLines == utw->scrollableLineIndex)
					{
						// We have reached the end of the scollable area
						utw->scrollLimit = 1;
						break;
					}
				}
			}
			else if (lines < 0)
			{
				// user wants to scroll the window down
				while (i > lines)
				{
					i--;
					
					if ((utw->viewableLineIndex + i) % utw->maxBufferLines == utw->scrollableLineIndex)
					{
						// We have reached the end of the scollable area
						utw->scrollLimit = 1;
						break;
					}
						
					if (utw->viewableLineIndex == 0)
						utw->viewableLineIndex = utw->maxBufferLines;
				}
			}
			
			// Now scroll the window up the specified amount
			utw->viewableLineIndex = (utw->viewableLineIndex += i) % utw->maxBufferLines;
		}
	}
	
	if (utw->scrollLimit)
	{
		// Inform the user tha twe have reached the end of the scrollable area.
		utw->scrollLimit = 0;
		return 1;
	}
		
	return 0;
}
